# -*- coding: utf-8 -*-

"""
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

from qgis.PyQt.QtCore import QCoreApplication
from qgis.core import (QgsProcessing,
                       QgsFeatureSink,
                       QgsProcessingException,
                       QgsProcessingAlgorithm,
                       QgsProcessingParameterFeatureSource,
                       QgsProcessingOutputVectorLayer,
                       QgsVectorFileWriter,
                       QgsProject,
                       QgsVectorLayer,
                       QgsTask,
                       QgsApplication,
                       QgsProcessingContext,
                       QgsMessageLog)
from qgis import processing
import numpy as np
import geopandas as gpd
from shapely.geometry import  Polygon, Point, LineString, MultiPoint, mapping
from shapely import geometry, ops
from shapely.ops import nearest_points
import pyproj
from geopandas import GeoSeries
from pyproj import CRS
import networkx as nx
from networkx import *
import fiona




class ExampleProcessingAlgorithm(QgsProcessingAlgorithm):
    """
    This is an example algorithm that takes a vector layer and
    creates a new identical one.

    It is meant to be used as an example of how to create your own
    algorithms and explain methods and variables used to do it. An
    algorithm like this will be available in all elements, and there
    is not need for additional work.

    All Processing algorithms should extend the QgsProcessingAlgorithm
    class.
    """

    # Constants used to refer to parameters and outputs. They will be
    # used when calling the algorithm from another algorithm, or when
    # calling from the QGIS console.

    INPUT_POL = 'INPUT_POL'
    INPUT_LI = 'INPUT_LI'

    def tr(self, string):
        """
        Returns a translatable string with the self.tr() function.
        """
        return QCoreApplication.translate('Processing', string)

    def createInstance(self):
        return ExampleProcessingAlgorithm()

    def name(self):
        """
        Returns the algorithm name, used for identifying the algorithm. This
        string should be fixed for the algorithm, and must not be localised.
        The name should be unique within each provider. Names should contain
        lowercase alphanumeric characters only and no spaces or other
        formatting characters.
        """
        return 'Schritt4_b_minimum_spanning_tree'

    def displayName(self):
        """
        Returns the translated algorithm name, which should be used for any
        user-visible display of the algorithm name.
        """
        return self.tr('Schritt4_b_minimum_spanning_tree')

    def group(self):
        """
        Returns the name of the group this algorithm belongs to. This string
        should be localised.
        """
        return self.tr('BBV Clustersuche')

    def groupId(self):
        """
        Returns the unique ID of the group this algorithm belongs to. This
        string should be fixed for the algorithm, and must not be localised.
        The group id should be unique within each provider. Group id should
        contain lowercase alphanumeric characters only and no spaces or other
        formatting characters.
        """
        return 'BBV Clustersuche'

    def shortHelpString(self):
        """
        Returns a localised short helper string for the algorithm. This string
        should provide a basic description about what the algorithm does and the
        parameters and outputs associated with it..
        """
        return self.tr("Example algorithm short description")

    def initAlgorithm(self, config=None):
        """
        Here we define the inputs and output of the algorithm, along
        with some other properties.
        """

        # We add the input vector features source. It can have any kind of
        # geometry.
        self.addParameter(
            QgsProcessingParameterFeatureSource(
                self.INPUT_POL,
                self.tr('Ausbaugebiete Polygon Layer'),
                [QgsProcessing.TypeVectorPolygon]
            )
        )
        
        self.addParameter(
            QgsProcessingParameterFeatureSource(
                self.INPUT_LI,
                self.tr('Wegnetz Line Layer'),
                [QgsProcessing.TypeVectorLine]
            )
        )

    def processAlgorithm(self, parameters, context, feedback):
        """
        Here is where the processing itself takes place.
        """

        po,pot = self.parameterAsCompatibleSourceLayerPathAndLayerName(parameters, self.INPUT_POL, context, QgsVectorFileWriter.supportedFormatExtensions()
                                                                                           ,'gpkg'  ,feedback=feedback)
        li,lit = self.parameterAsCompatibleSourceLayerPathAndLayerName(parameters, self.INPUT_LI, context, QgsVectorFileWriter.supportedFormatExtensions()
                                                                                           ,'shp'   ,feedback=feedback)
    #----------------------------------------------------------------------------------------------------------------
        line = gpd.read_file(li)
        line_crs = line.crs
        c = fiona.open(po, layer=pot)
        pol = gpd.GeoDataFrame.from_features([feature for feature in c], crs=line_crs)
        

        line.drop(line.loc[line['networkGrp']!=0].index, inplace=True)
        line = line.reset_index(drop=True)
        line['length'] = line['geometry'].length.round(2)
        line['order'] = line.index
    #---------------------------------------------------------------------------
        G = nx.Graph()
        G.graph['crs'] = line.crs
        fields = list(line.columns)

        for index, row in line.iterrows():
            first = row.geometry.coords[0]
            last = row.geometry.coords[-1]

            data = [row[f] for f in fields]
            attributes = dict(zip(fields, data))
            G.add_edge(first, last, **attributes)

        first = list(G.edges(data=True))


        weighted_G = nx.Graph()
        for data in first:
            weighted_G.add_edge(data[0],data[1], order = data[2]['order'] , weight=data[2]['length'])

        edge_labels = nx.get_edge_attributes(weighted_G, 'order')
        pos = {v:v for v in weighted_G.nodes()}

        nodes= list(weighted_G.nodes)

        node_xy, node_data = zip(*weighted_G.nodes(data=True))
        point = gpd.GeoDataFrame(list(node_data), geometry=[Point(i, j) for i, j in node_xy])
        point.crs = CRS.from_epsg(25832).to_wkt()
        point = point.to_crs("EPSG:25832")


#--------------------------------------------------------------------------
        point['pol_no'] = -9999
        for i in range(len(point)):
            for j in range(len(pol)):
                if point.loc[i,'geometry'].within(pol.loc[j,'geometry']):
                    point.loc[i,'pol_no']= j

#--------------------------------------------------------------------------
        tree = nx.Graph()
        pdict = {}
        
         
        for ii in range(len(pol)):
            for jj in range(len(pol)):
                if ii < jj:
                    pol_first = pol.iloc[[ii]]
                    pol_second = pol.iloc[[jj]]

                    g = [i for i in pol_first.geometry]
                    first_coords = mapping(g[0])["coordinates"]
                    first_coords = MultiPoint (first_coords[0])

                    g = [i for i in pol_second.geometry]
                    second_coords = mapping(g[0])["coordinates"]
                    second_coords = MultiPoint (second_coords[0])

                    nearest_geoms = nearest_points(first_coords, second_coords)
                    nearest_first = nearest_geoms[0]
                    nearest_second = nearest_geoms[1]


                    short_eu = LineString ([nearest_first, nearest_second])
                    mid_point = short_eu.interpolate(0.5, normalized = True)

           #-----------------------------------------------------------------------------

                    top_p = 3
                    top_od =[]

                    top_0 = point.copy()
                    top_0.drop(top_0.loc[top_0['pol_no']!=ii].index, inplace=True)
                    top_0['dis_mid'] = top_0['geometry'].distance(mid_point)
                    top_0 = top_0.nsmallest(top_p, 'dis_mid')
                    top_od.append(top_0.index.tolist())

                    top_1 = point.copy()
                    top_1.drop(top_1.loc[top_1['pol_no']!=jj].index, inplace=True)
                    top_1['dis_mid'] = top_1['geometry'].distance(mid_point)
                    top_1 = top_1.nsmallest(top_p, 'dis_mid')
                    top_od.append(top_1.index.tolist())

                    length_set = []
                    short_path_set = []

                    for a in range(len(top_od[0])):
                        for b in range(len(top_od[1])):
                            ori = top_od [0][a]
                            des = top_od [1][b]
                            short_p = nx.astar_path(weighted_G,source=nodes[ori],target=nodes[des],weight='weight')
                            path_edges = [edge_labels.get(x, edge_labels.get((x[1],x[0]))) for x in zip(short_p, short_p[1:])]
                            short_path_set.append(path_edges)
                            short_length = nx.astar_path_length(weighted_G,source=nodes[ori],target=nodes[des],weight='weight')
                            length_set.append(short_length)

                    shortest_idx = length_set.index(min(length_set))
                    best_path = short_path_set[shortest_idx]
                    leng = min(length_set)
                    
                    pdict [ii,jj] = best_path
                    tree.add_edge(ii, jj, weight=leng)
            

#-----------------------------------------------------------------------------
        best_route=[]
        mst = list(nx.minimum_spanning_edges(tree))
        for i in range(len(mst)):
            strt = mst [i][0]
            end = mst [i][1]
            best_route.extend(pdict[strt,end])
        
        all_path = line.iloc[best_route]
        #vl = QgsVectorLayer(all_pat,"mygeojson","ogr")
        #v2 = QgsVectorLayer(pol,"eojson","ogr")
        #v1.setCrs(line_crs)
        #v2.setCrs(line_crs)
        #param = {'INPUT' : all_pat, 'OVERLAY' : self.INPUT_POL, 'OUTPUT' : 'memory'}
        #differenced = processing.run('qgis:difference', param)
        #all_path = gpd.GeoDataFrame (data = differenced['OUTPUT'])
        
        
        #all_path = gpd.overlay(all_path, pol, how='difference')
        #all_path['length'] = all_path['geometry'].length.round(2)
    
        projectdir = QgsProject.instance().readPath("./")
        projectname = QgsProject.instance().baseName()
        output_dir = projectdir+'/'+projectname+'.gpkg'
        all_path.to_file(output_dir, layer='mst', driver="GPKG")

        dispaly_dir= output_dir+"|layername=mst"


        context.addLayerToLoadOnCompletion(dispaly_dir, QgsProcessingContext.LayerDetails('first result layer', QgsProject.instance(), ''))

        results = {}
        return results

    
        
