import numpy as np
import geopandas as gpd
from shapely.geometry import  Polygon, Point, LineString, MultiPoint, mapping
from shapely import geometry, ops
from shapely.ops import nearest_points
import pyproj
import matplotlib.pyplot as plt
from geopandas import GeoSeries
from pyproj import CRS
import networkx as nx
from networkx import *
from datetime import datetime
import fiona
startTime = datetime.now()
from joblib import Parallel, delayed
import multiprocessing


#def short_tree(line, polygon):                                         
line = gpd.read_file('wegnetz4.shp')
#line = gdf_line
#line_crs = line.crs
#c = fiona.open('C:/Users/Mostafa/Downloads/test_mittwoch.gpkg', \
    #           layer='Ausbaugebiete')
#pol = gpd.GeoDataFrame.from_features([feature for feature in c], crs=line_crs)
pol = gpd.read_file('ausbau4.shp')
#pol = gdf_polygon
line.drop(line.loc[line['networkGrp']!=0].index, inplace=True)
line = line.reset_index(drop=True)
line['length'] = line['geometry'].length.round(2)
line['order'] = line.index
print(datetime.now() - startTime)
#---------------------------------------------------------------------------
G = nx.Graph()
G.graph['crs'] = line.crs
fields = list(line.columns)

for index, row in line.iterrows():
   first = row.geometry.coords[0]
   last = row.geometry.coords[-1]

   data = [row[f] for f in fields]
   attributes = dict(zip(fields, data))
   G.add_edge(first, last, **attributes)

first = list(G.edges(data=True))


weighted_G = nx.Graph()
for data in first:
   weighted_G.add_edge(data[0],data[1], order = data[2]['order'] , weight=data[2]['length'])

edge_labels = nx.get_edge_attributes(weighted_G, 'order')
pos = {v:v for v in weighted_G.nodes()}

nodes= list(weighted_G.nodes)

node_xy, node_data = zip(*weighted_G.nodes(data=True))
point = gpd.GeoDataFrame(list(node_data), geometry=[Point(i, j) for i, j in node_xy])
point.crs = CRS.from_epsg(25832).to_wkt()
point = point.to_crs("EPSG:25832")

print(datetime.now() - startTime)

#--------------------------------------------------------------------------
point['pol_no'] = -9999
for i in range(len(point)):
   for j in range(len(pol)):
      if point.loc[i,'geometry'].within(pol.loc[j,'geometry']):
         point.loc[i,'pol_no']= j

print(datetime.now() - startTime)
#--------------------------------------------------------------------------

def short_tree(ii, jj): 
   pol_first = pol.iloc[[ii]]
   pol_second = pol.iloc[[jj]]

   g = [i for i in pol_first.geometry]
   first_coords = mapping(g[0])["coordinates"]
   first_coords = MultiPoint (first_coords[0])

   g = [i for i in pol_second.geometry]
   second_coords = mapping(g[0])["coordinates"]
   second_coords = MultiPoint (second_coords[0])

   nearest_geoms = nearest_points(first_coords, second_coords)
   nearest_first = nearest_geoms[0]
   nearest_second = nearest_geoms[1]


   short_eu = LineString ([nearest_first, nearest_second])
   mid_point = short_eu.interpolate(0.5, normalized = True)

   #-----------------------------------------------------------------------------

   top_p = 3
   top_od =[]

   top_0 = point.copy()
   top_0.drop(top_0.loc[top_0['pol_no']!=ii].index, inplace=True)
   top_0['dis_mid'] = top_0['geometry'].distance(mid_point)
   top_0 = top_0.nsmallest(top_p, 'dis_mid')
   top_od.append(top_0.index.tolist())

   top_1 = point.copy()
   top_1.drop(top_1.loc[top_1['pol_no']!=jj].index, inplace=True)
   top_1['dis_mid'] = top_1['geometry'].distance(mid_point)
   top_1 = top_1.nsmallest(top_p, 'dis_mid')
   top_od.append(top_1.index.tolist())

   length_set = []
   short_path_set = []

   for a in range(len(top_od[0])):
       for b in range(len(top_od[1])):
           ori = top_od [0][a]
           des = top_od [1][b]
           short_p = nx.astar_path(weighted_G,source=nodes[ori],target=nodes[des],weight='weight')         
           path_edges = [edge_labels.get(x, edge_labels.get((x[1],x[0]))) for x in zip(short_p, short_p[1:])]
           short_path_set.append(path_edges)
           short_length = nx.astar_path_length(weighted_G,source=nodes[ori],target=nodes[des],weight='weight')
           length_set.append(short_length)

   shortest_idx = length_set.index(min(length_set))
   best_path = short_path_set[shortest_idx]
   leng = min(length_set)
   return ii ,jj, best_path, leng 

num_cores = multiprocessing.cpu_count()
par = Parallel(n_jobs=num_cores, verbose=50) \
  (delayed(short_tree)(ii,jj) for ii in range(len(pol)) for jj in range(len(pol)) if ii<jj)

print(datetime.now() - startTime)
par = [i for i in par if i != None]

tree = nx.Graph()
pdict = {}

for i in range(len(par)):
    tree.add_edge(par[i][0], par[i][1], weight=par[i][3])
    pdict [par[i][0],par[i][1]] = par[i][2]


#-----------------------------------------------------------------------------
best_route=[]
mst = list(nx.minimum_spanning_edges(tree))
for i in range(len(mst)):
    strt = mst [i][0]
    end = mst [i][1]
    best_route.extend(pdict[strt,end])
   
all_path = line.iloc[best_route]
all_path = gpd.overlay(all_path, pol, how='difference')
all_path['length'] = all_path['geometry'].length.round(2)
#-----------------------------------------------------------------------------
line.plot(ax=plt.gca(), color='black', alpha=0.5)
#point.plot(ax=plt.gca(), color='blue', alpha=0.7)
all_path.plot(ax=plt.gca(), color='red', alpha=1)
pol.plot(ax=plt.gca(), color='green', alpha=0.5)
all_path.to_file("mst.shp")

#plt.show()


